from typing import (
    Optional,
    Type,
)

from click import ClickException
from pydantic import ValidationError
from pydantic.error_wrappers import display_errors


__all__ = [
    "TransitionExecutionError",
    "ConfigValidationError",
    "StatemachineFactoryLoadError",
    "StatemachineFactoryTypeError",
    "SkipSectionError",
]


class TransitionExecutionError(Exception):
    """Base class for errors that occur during state transitions"""

    def __init__(
        self,
        message: str,
        cause: Optional[Exception] = None,
        fallback_state: Optional[str] = None,
    ):
        """
        Args:
            message: The error message
            cause: The underlying cause of the transition error.
            fallback_state: Optionally the name of the state to fallback to.
        """
        super().__init__(message)
        self.cause = cause
        self.fallback_state = fallback_state

    def __str__(self):
        ret = f"message: '{super().__str__()}'"
        if self.cause:
            ret += f" cause: {self.cause}"
        return ret


class ConfigValidationError(ClickException):
    """CLI exception indicating that the configuration file was not valid"""

    def __init__(self, cause: ValidationError):
        """
        Args:
            cause: Pydantic validation error that caused validation to fail.
        """
        formated_errors = display_errors(cause.errors())
        super().__init__(
            f"Failed to validate the configuration file.\n{formated_errors}"
        )
        self.__cause__ = cause


class StatemachineFactoryLoadError(ClickException):
    """CLI Exception indicating that a sm factory plugin could not be loaded"""

    def __init__(self, factory_name: str):
        """
        Args:
            factory_name: The name of the factory that failed to load
        """
        super().__init__(message=f"Failed to load sm factory: '{factory_name}'")


class StatemachineFactoryTypeError(ClickException):
    """CLI Exception indicating that a type error occurred while loading a sm factory plugin."""

    def __init__(self, factory_type: Type):
        """
        Args:
            factory_type: The python type of the factory that could not be loaded.
        """
        super().__init__(
            message=f"Failed to load sm factory plugin got invalid type: '{factory_type}'"
        )


class SkipSectionError(KeyboardInterrupt):
    """Utility exception to indicate that a task section should be skipped"""



import sys

from importlib.util import (
    module_from_spec,
    spec_from_file_location,
)
from pathlib import Path
from typing import (
    Dict,
    List,
    Type,
)

from . import FACTORY_ENTRYPOINT
from .config import PluginConfig
from .errors import (
    StatemachineFactoryLoadError,
    StatemachineFactoryTypeError,
)
from .sm import StatemachineFactory


try:
    from importlib.metadata import EntryPoint  # type: ignore
    from importlib.metadata import entry_points  # type: ignore
except ImportError:
    # need to use backport for python < 3.8
    from importlib_metadata import (
        EntryPoint,
        entry_points,
    )


def get_factories(plugin_config: PluginConfig) -> Dict[str, EntryPoint]:
    # all available factories which are also included as per the config
    available_sm_factories: List[EntryPoint] = [
        ep
        for ep in entry_points().get(FACTORY_ENTRYPOINT, [])
        if any([pattern.match(ep.name) for pattern in plugin_config.include_names])
    ]

    # filter out plugins explicitly excluded
    allowed_sm_factories: Dict[str, EntryPoint] = {
        ep.name: ep
        for ep in available_sm_factories
        if not any([pattern.match(ep.name) for pattern in plugin_config.exclude_names])
    }

    return allowed_sm_factories


def get_factory(
    available_sm_factories: Dict[str, EntryPoint], sm_name: str
) -> StatemachineFactory:
    try:
        # try to load from python file
        if sm_name.endswith(".py"):
            # check if file exists
            plugin_path = Path(sm_name)
            if plugin_path.exists():
                # load the plugin file
                spec = spec_from_file_location(plugin_path.stem, plugin_path)
                factory_plugin = module_from_spec(spec)
                sys.modules[plugin_path.stem] = factory_plugin
                spec.loader.exec_module(factory_plugin)  # type: ignore
                # get the plugin class (must be name StateMachineFactory)
                factory_class: Type[
                    StatemachineFactory
                ] = factory_plugin.StatemachineFactory  # type: ignore
            else:
                raise StatemachineFactoryLoadError(sm_name)

        # try to load from entrypoint
        else:
            sm_entrypoint: EntryPoint = available_sm_factories[sm_name]
            factory_class = sm_entrypoint.load()

        if issubclass(factory_class, StatemachineFactory):
            return factory_class()
        else:
            raise StatemachineFactoryTypeError(factory_class)
    except KeyError:
        raise StatemachineFactoryLoadError(sm_name)



import random
import re

from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Pattern,
    Type,
)

import numpy as np

from pydantic import (
    BaseModel,
    BaseSettings,
    Field,
    ValidationError,
    validator,
)
from ruamel.yaml import YAML

from .errors import ConfigValidationError
from .model import (
    LogLevel,
    StatemachineConfig,
)


__all__ = [
    "PluginConfig",
    "LoggingConfig",
    "LogFormat",
    "LogHandler",
    "FileLogHandler",
    "Settings",
    "load_config_file",
    "load_sm_config",
    "load_settings",
    "get_seed",
    "configure_seed",
]

_SEED: int


class PluginConfig(BaseModel):
    """Configuration options for the state machine factory plugin system."""

    include_names: List[Pattern] = Field(
        [re.compile(r".*")],
        description="A list of regular expressions used to define which plugins to include.",
    )
    exclude_names: List[Pattern] = Field(
        [],
        description="A list of regular expressions used to define \
        which plugins to explicitly exclude.",
    )


class LogFormat(str, Enum):
    """Enum for log formatter styles"""

    PLAIN = "plain"
    """Human readable structured text format"""

    COLORED = "colored"
    """The same as PLAIN, but colorized"""

    JSON = "json"
    """The log events in JSON format"""


class LogHandler(BaseModel):
    """Configuration for a log handler"""

    enabled: bool = Field(
        False,
        description="If the log handler should be enabled or not",
    )
    format: LogFormat = Field(
        LogFormat.PLAIN,
        description="The log format to use when logging to the console",
    )


class FileLogHandler(LogHandler):
    """Configuration for a file log handler"""

    format: LogFormat = Field(
        LogFormat.JSON,
        description="The log format to use when logging to the console",
    )

    path: Path = Field(
        Path("sm.log"),
        description="The file path to log to, must be set if the handler is enabled.",
    )

    @validator("path")
    def validate_log_path(cls, path: Path) -> Path:
        """Validate that the given path is a file if it exists"""
        if path.exists() and not path.is_file():
            raise ValueError(f"Log file path {path.absolute()} is not a file!")
        return path


class LogTimestamp(BaseModel):
    """Configuration for the log timestamp format and key"""

    format: Optional[str] = Field(
        None,
        description=(
            r"The strftime string format to use for formating timestamps (e.g., `%Y-%m-%d %H:%M:%S`). "
            "If this is `None` a [UNIX timestamp](https://en.wikipedia.org/wiki/Unix_time) is used."
        ),
    )
    utc: bool = Field(
        True,
        description="If the timestamp should be in UTC or the local timezone.",
    )
    key: str = Field(
        "timestamp",
        description="The key to use for the timestamp.",
    )


class LoggingConfig(BaseSettings):
    """Configuration options for the logging system."""

    level: LogLevel = Field(
        LogLevel.WARNING,
        description="The log level to use for logging",
    )

    timestamp: LogTimestamp = Field(
        LogTimestamp(),
        description="Configuration options for modifying the log timestamps",
    )

    console: LogHandler = Field(
        LogHandler(enabled=True, format=LogFormat.COLORED),
        description="Configuration for the console logger",
    )

    file: FileLogHandler = Field(
        FileLogHandler(enabled=False, format=LogFormat.JSON),
        description="Configuration for the file logger",
    )


class Settings(BaseSettings):
    """Cyber Range Kyoushi Simulation settings"""

    seed: Optional[int] = Field(
        None,
        description="The seed to use for random generators",
    )

    plugin: PluginConfig = Field(
        PluginConfig(),
        description="The plugin system configuration",
    )

    log: LoggingConfig = Field(
        LoggingConfig(),
        description="The logging configuration",
    )


def load_config_file(config_path: Path) -> Dict[Any, Any]:
    """Loads a given a config from the given path and returns a raw dictionary.

    Supported file formats are:
        - YAML

    Args:
        config_path: The file path to read the config from

    Returns:
        The contents of the configuration file converted to a dictionary or `{}`
        if the file is empty or does not exist.
    """
    yaml = YAML(typ="safe")
    if config_path.exists():
        return yaml.load(config_path)
    return {}


def load_sm_config(
    config_path: Path,
    sm_config_type: Type[StatemachineConfig],
) -> StatemachineConfig:
    """Loads and validates the state machine configuration

    Args:
        config_path: The path to load the configuration from
        sm_config_type: The state machine config class to convert to

    Raises:
        ConfigValidationError: If the config validation fails

    Returns:
        The state machine configuration validated and converted
        to the given config class.
    """
    try:
        config_raw = load_config_file(config_path)
        if issubclass(sm_config_type, BaseModel):
            return sm_config_type.parse_obj(config_raw)
        else:
            return config_raw

    except ValidationError as val_err:
        raise ConfigValidationError(val_err)


def load_settings(
    settings_path: Path,
    log_level: Optional[LogLevel] = None,
    seed: Optional[int] = None,
) -> Settings:
    """Loads the Cyber Range Kyoushi Simulation CLI settings

    Args:
        settings_path: The path to load the settings file from
        log_level: The CLI log_level override. This supercedes the config
                   and environment variable values..

    Raises:
        ConfigValidationError: If the config validation fails

    Returns:
        The validated settings object
    """
    try:
        settings_raw = load_config_file(settings_path)

        if log_level is not None:
            settings_raw.setdefault("log", {})["level"] = log_level

        if seed is not None:
            settings_raw["seed"] = seed

        settings = Settings(
            _env_file=str(object),
            _env_file_encoding=None,
            _secrets_dir=None,
            **settings_raw,
        )

        return settings
    except ValidationError as val_err:
        raise ConfigValidationError(val_err)


def configure_seed(seed: Optional[int] = None):
    """Configure a seed for PRNG, if no seed is passed then one is generated.

    Args:
        seed: The seed to use for PRNG
    """
    if seed is None:
        # legacy global numpy requires seeds to be in
        # [0, 2**32-1]
        seed = random.randint(0, 2 ** 32 - 1)
    global _SEED
    _SEED = seed
    random.seed(seed)
    np.random.seed(seed)


def get_seed() -> int:
    """Get the global random seed value for the simulation library

    Returns:
        The seed value
    """
    # configure seed if not already done
    global _SEED
    if _SEED is None:
        configure_seed()

    # return seed value
    return _SEED



import json
import logging
import logging.config
import re

from functools import partial
from typing import (
    Any,
    Callable,
    List,
    Mapping,
    MutableMapping,
    Tuple,
    Union,
)

import structlog

from pydantic.json import custom_pydantic_encoder

from .config import (
    LogFormat,
    LoggingConfig,
)


JSON_ENCODERS = {
    re.Pattern: lambda p: p.pattern,
}


encoder = partial(custom_pydantic_encoder, JSON_ENCODERS)
"""
Custom encoder based on the pydantic encoder with additional types enabled.
(see `JSON_ENCODERS`)

!!! Note
    This is necessary since the Pydantic JSON encoder
    cannot handle `Pattern` fields as of v1.7.3. The code exists already on
    the master branch and should be part of the next release.
"""


LOGGER_NAME = "cr_kyoushi.simulation"
"""The name of the Cyber Range Kyoushi Simulation logger"""


def rename_event_key(
    logger: logging.Logger,
    method_name: str,
    event_dict: MutableMapping[str, Any],
    field_name: str,
) -> MutableMapping[str, Any]:
    """Processor for renaming the event key to message.

    We do this since this is more compatible with the ecs schema.
    """
    event_dict[field_name] = event_dict.pop("event")
    return event_dict


def rename_event_key_wrapper(
    func: Callable[[logging.Logger, str, MutableMapping[str, Any]], Any],
    field_name: str = "message",
) -> Callable[
    [logging.Logger, str, MutableMapping[str, Any]], MutableMapping[str, Any]
]:
    """Processor wrapper that renames the event field before calling the given processor.

    Args:
        func: The processor to be wrapped
        field_name: The field name to use instead of `event`

    Returns:
        The wrapped processor
    """

    def _rename_event_key(
        logger: logging.Logger, method_name: str, event_dict: MutableMapping[str, Any]
    ) -> MutableMapping[str, Any]:
        event_dict = rename_event_key(logger, method_name, event_dict, field_name)
        return func(logger, method_name, event_dict)

    return _rename_event_key


def get_logger() -> structlog.stdlib.BoundLogger:
    """Convenience function for getting the Cyber Range Kyoushi Simulation logger."""
    return structlog.stdlib.get_logger(LOGGER_NAME)


def configure_logging(logging_config: LoggingConfig):
    """Configures the logging system based on the passed
    [`LoggingConfig`][cr_kyoushi.simulation.config.LoggingConfig]

    Args:
        logging_config: The logging configuration object
    """
    # ensure we start from default config
    structlog.reset_defaults()

    timestamper = structlog.processors.TimeStamper(
        fmt=logging_config.timestamp.format,
        utc=logging_config.timestamp.utc,
        key=logging_config.timestamp.key,
    )
    # shared processors for standard lib and structlog
    shared_processors: List[
        Callable[
            [Any, str, MutableMapping[str, Any]],
            Union[Mapping[str, Any], str, bytes, Tuple[Any, ...]],
        ]
    ] = [
        structlog.stdlib.add_log_level,
        timestamper,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
    ]

    # processor only for structlog
    processors: List[
        Callable[
            [Any, str, MutableMapping[str, Any]],
            Union[Mapping[str, Any], str, bytes, Tuple[Any, ...]],
        ]
    ] = [structlog.stdlib.filter_by_level]
    processors.extend(shared_processors)
    # the processor formatter wrapper must be last as it changes the
    processors.append(structlog.stdlib.ProcessorFormatter.wrap_for_formatter)

    handlers = {}
    # configure console logging
    if logging_config.console.enabled:
        handlers["console"] = {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": logging_config.console.format,
        }

    # configure file logging
    if logging_config.file.enabled:
        handlers["file"] = {
            "level": "DEBUG",
            "class": "logging.handlers.WatchedFileHandler",
            "filename": str(logging_config.file.path.absolute()),
            "formatter": logging_config.file.format,
        }

    # configure standard lib logging
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                LogFormat.PLAIN: {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processor": structlog.dev.ConsoleRenderer(colors=False),
                    "foreign_pre_chain": shared_processors,
                },
                LogFormat.COLORED: {
                    "()": structlog.stdlib.ProcessorFormatter,
                    "processor": structlog.dev.ConsoleRenderer(colors=True),
                    "foreign_pre_chain": shared_processors,
                },
                LogFormat.JSON: {
                    "()": structlog.stdlib.ProcessorFormatter,
                    # when logging to json we rename the "event" field to "message"
                    "processor": rename_event_key_wrapper(
                        structlog.processors.JSONRenderer(
                            serializer=json.dumps,
                            sort_keys=True,
                            default=encoder,
                        )
                    ),
                    "foreign_pre_chain": shared_processors,
                },
            },
            "handlers": handlers,
            "loggers": {LOGGER_NAME: {"handlers": handlers.keys(), "propagate": True}},
        }
    )

    # apply structlog config
    structlog.configure(
        processors=processors,
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    # set the log level
    logger = structlog.get_logger(LOGGER_NAME)
    logger.setLevel(logging_config.level)



import logging
import random

from datetime import (
    datetime,
    time,
    timedelta,
)
from enum import IntEnum
from typing import (
    Any,
    Dict,
    Optional,
    Set,
    TypeVar,
    Union,
)

from pydantic import (
    BaseModel,
    Field,
    validator,
)


__all__ = [
    "StatemachineConfig",
    "Context",
    "Weekday",
    "TimePeriod",
    "WeekdayActivePeriod",
    "ComplexActivePeriod",
    "SimpleActivePeriod",
    "ActivePeriod",
    "ApproximateFloat",
]

StatemachineConfig = TypeVar("StatemachineConfig", Dict[Any, Any], BaseModel)
"""Placeholder generic type for state machine configurations."""

Context = Union[BaseModel, Dict[str, Any]]
"""
    Contexts are used to store the state machine execution context.

    They can either be a custom defined pydantic model or `Dict[str, Any]`.
"""


class LogLevel(IntEnum):
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

    @classmethod
    def lookup(cls):
        return {v: k for v, k in cls.__members__.items()}

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Union[str, int, "LogLevel"]):
        """Parses and validates `LogLevel` input in either `str`, `int` or `Enum` encoding.

        Args:
            val: The encoded input log level

        Raises:
            ValueError: if the given input is not a valid log level

        Returns:
            LogLevel enum
        """
        # check enum input
        if isinstance(val, LogLevel):
            return val
        # check int LogLevel input
        if isinstance(val, int):
            if val in cls.lookup().values():
                return LogLevel(val)
            raise ValueError("invalid integer LogLevel")
        # check str LogLevel input
        try:
            return cls.lookup()[val.upper()]
        except KeyError as key_error:
            raise ValueError("invalid string LogLevel") from key_error


class Weekday(IntEnum):
    """
    Enumeration for representing the days of the week.

    Weekdays are represented as the integers 0-6 and can
    be constructed from either their int representations or
    their english names.
    """

    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6

    @classmethod
    def lookup(cls):
        return {v: k for v, k in cls.__members__.items()}

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, val: Union[str, int, "Weekday"]):
        """Parses and validates `Weekday` input in either `str`, `int` or `Enum` encoding.

        Args:
            val: The encoded input week day

        Raises:
            ValueError: if the given input is not a valid weekday

        Returns:
            Weekday enum
        """
        # check enum input
        if isinstance(val, Weekday):
            return val
        # check int weekday input
        if isinstance(val, int):
            if val in cls.lookup().values():
                return Weekday(val)
            raise ValueError("invalid integer weekday")
        # check str weekday input
        try:
            return cls.lookup()[val.upper()]
        except KeyError as key_error:
            raise ValueError("invalid string weekday") from key_error


class TimePeriod(BaseModel):
    """A time period as defined by a start and end time."""

    start_time: time = Field(description="The start time of the period")
    end_time: time = Field(description="The end time of the period")

    def in_period(self, to_check: time) -> bool:
        """Checks wether the given time of the day is within the scope of this time period.

        Args:
            to_check (time): The time of the day to check

        Returns:
            bool: `True` if within the time period `False` otherwise
        """
        if self.start_time <= self.end_time:
            return self.start_time <= to_check and self.end_time > to_check
        # start > end means our time period is between two days
        return self.start_time <= to_check or self.end_time > to_check


class WeekdayActivePeriod(BaseModel):
    """
    A `WeekdayActivePeriod` defines its active time based on the configure week day
    and time period.
    """

    week_day: Weekday = Field(description="The active day of the week")
    time_period: Optional[TimePeriod] = Field(
        description="The active time period (if this is not set the whole day is considered active)"
    )

    def in_active_period(self, to_check: datetime) -> bool:
        """Checks wether the given datetime is within the scope of this active period.

           This will be True if the week day matches and given time of the day is
           within the time period.

        Args:
            to_check (datetime): The datetime to check

        Returns:
            bool: `True` if inside the active period `False` otherwise
        """
        return Weekday(to_check.weekday()) is self.week_day and (
            self.time_period is None or self.time_period.in_period(to_check.time())
        )

    def __hash__(self):
        return hash(self.week_day)

    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__ and self.week_day == other.week_day


class ComplexActivePeriod(BaseModel):
    """
    A `ComplexActivePeriod` is defined by a set of
    [`WeekdayActivePeriod`][cr_kyoushi.simulation.model.WeekdayActivePeriod].

    This makes it possible to configure active times for each week day separately.
    """

    week_days: Set[WeekdayActivePeriod] = Field(
        description="Set of active periods, each week day can only have one configuration"
    )

    def in_active_period(self, to_check: datetime) -> bool:
        """Checks wether the given datetime is within this active period.

           This is `True` if the datetime is within the scope of **one** of
           the [`WeekdayActivePeriods`][cr_kyoushi.simulation.model.WeekdayActivePeriod].


        Args:
            to_check (datetime): The datetime to check

        Returns:
            bool: `True` if inside the active period `False` otherwise
        """
        return any(period.in_active_period(to_check) for period in self.week_days)


class SimpleActivePeriod(BaseModel):
    """
    Similar to [`ComplexActivePeriod`][cr_kyoushi.simulation.model.ComplexActivePeriod],
    but each week day has the same active time period.
    """

    week_days: Set[Weekday] = Field(description="A set of active week days.")
    time_period: Optional[TimePeriod] = Field(
        description="The daylie active time period \
        (if this is not set the whole days are considered active)."
    )

    def in_active_period(self, to_check: datetime) -> bool:
        """Checks wether the given datetime is within this active period.

           This is `True` if the datetime matches **one** of the active week days
           and the active time period.


        Args:
            to_check (datetime): The datetime to check

        Returns:
            bool: `True` if inside the active period `False` otherwise
        """
        return Weekday(to_check.weekday()) in self.week_days and (
            self.time_period is None or self.time_period.in_period(to_check.time())
        )


class ActivePeriod(BaseModel):
    """Active period union type.

    This class can be used in pydantic models to allow
    loading of [`ComplexActivePeriod`][cr_kyoushi.simulation.model.ComplexActivePeriod]
    or [`SimpleActivePeriod`][cr_kyoushi.simulation.model.SimpleActivePeriod].
    """

    __root__: Union[ComplexActivePeriod, SimpleActivePeriod]

    def in_active_period(self, to_check: datetime) -> bool:
        """Checks wether the given datetime is within this active period.

        The actual check is delegated to the underlying sub types own
        checking logic.

        Args:
            to_check (datetime): The datetime to check

        Returns:
            bool: `True` if inside the active period `False` otherwise
        """
        return self.__root__.in_active_period(to_check)


class ApproximateFloat(BaseModel):
    """Approximate float value within a given open interval"""

    min: float = Field(description="The lower boundary for the float value")
    max: float = Field(description="The upper boundary for the float value")

    @validator("max")
    def validate_min_le_max(cls, v: float, values, **kwargs) -> float:
        """Custom validator to ensure min <= max"""
        assert values["min"] <= v, "Invalid boundaries must be min <= max"
        return v

    @classmethod
    def convert(cls, value: float) -> "ApproximateFloat":
        """Coerces a single float value to its ApproximateFloat equivalent.

        i.e., value = min = max

        Args:
            value: The float value to convert

        Returns:
            `Approximate(min=value, max=value)`
        """
        return ApproximateFloat(min=value, max=value)

    @property
    def value(self) -> float:
        """Gets a random float within the approximate float range

        Returns:
            A float x for which `min <= x <= max`
        """
        return random.uniform(self.min, self.max)


class WorkHours(TimePeriod):
    """A special type of time period that does not allow over night periods"""

    @validator("end_time")
    def validate_start_before_end(
        cls,
        v: time,
        values: Dict[str, Any],
        **kwargs,
    ) -> time:
        """Validates that the start time is before the end time.

        This restricts work hours to be within a single day i.e., work hours
        can not be defined to start on one day and end on the following day.

        Args:
            v: The parsed end time to check
            values: A dictionary containing previously parsed and validated fields.
                    See [Pydantic](https://pydantic-docs.helpmanual.io/usage/validators/) for details.

        Returns:
            The end time if it is valid
        """
        # if start time is not in values it failed to validate
        # so we skip the check and let validation fail
        if "start_time" in values:
            assert values["start_time"] < v, "End time must be after start time"

        return v


class WorkSchedule(BaseModel):
    """A weekly work schedule represented by the week days and work hours for each day"""

    work_days: Dict[Weekday, WorkHours] = Field(
        description="Dictionary containing the work hours for each weekday"
    )

    def is_work_day(self, weekday: Weekday) -> bool:
        """Checks wether the given weekday is a work day.

        Args:
            weekday: The weekday to check encoded as integer (0-6)

        Returns:
            `True` if it is a work day `False` otherwise
        """
        return weekday in self.work_days

    def is_work_time(self, to_check: datetime) -> bool:
        """Checks wether the given datetime is work time.

        Something is considered to be work time if it is a work day
        and the time is within the work hours of that work day.

        Args:
            to_check: The datetime to check

        Returns:
            `True` if the given datetime is work time `False` otherwise
        """
        weekday: Weekday = Weekday(to_check.weekday())
        # check if the datetime is on a work day
        if self.is_work_day(weekday):
            # if its on a workday check if its with the days work hours
            return self.work_days[weekday].in_period(to_check.time())

        return False

    def next_work_start(self, to_check: datetime) -> Optional[datetime]:
        """Gets the next work time, relative to the given datetime.

        If the given datetime is within work hours the start time
        of that work day is returned.

        Args:
            to_check: The datetime to find the next work time for

        Returns:
            The next work time or `None` if there is no work time
        """

        weekday: Weekday = Weekday(to_check.weekday())

        if (
            # if the given datetime is a workday
            self.is_work_day(weekday)
            and (
                # and work has not begun yet
                to_check.time() <= self.work_days[weekday].start_time
                # or if we are still within work time
                # we return the given days start time
                or self.is_work_time(to_check)
            )
        ):
            return datetime.combine(to_check.date(), self.work_days[weekday].start_time)

        # otherwise next work start must be some day after
        # the given day so we check the next 7 days
        # (we might only work once a week)
        for i in range(1, 8):
            # increment weekday and be sure to start from 0 once we pass sunday (int: 6)
            weekday = Weekday((weekday + 1) % 7)
            if self.is_work_day(weekday):
                # add the days till the next work day to the given date
                next_work: datetime = to_check + timedelta(i)
                # get the start time for that day
                start_time = self.work_days[weekday].start_time

                return datetime.combine(next_work.date(), start_time)

        # if we got here then no workday is set so we will never start
        return None



from abc import (
    ABCMeta,
    abstractmethod,
)
from itertools import cycle
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
)

import numpy as np

from structlog.stdlib import BoundLogger

from .model import Context
from .transitions import Transition
from .util import calculate_propabilities


__all__ = [
    "State",
    "SequentialState",
    "FinalState",
    "RoundRobinState",
    "ProbabilisticState",
    "EquallyRandomState",
]


class State(metaclass=ABCMeta):
    """A State contains various transitions to other states"""

    @property
    def name(self) -> str:
        """The name (including the prefix) of the state instance

        Names must be unique within a state machine.
        """
        if self._name_prefix is not None:
            return f"{self._name_prefix}_{self._name}"
        return self._name

    @property
    def name_only(self) -> str:
        """The name of the state instance (names must be uniq within a state machine)."""
        return self._name

    @property
    def name_prefix(self) -> Optional[str]:
        """The name prefix of the state instance."""
        return self._name_prefix

    @property
    def transitions(self) -> List[Transition]:
        """
        List of all possible [`transitions`][cr_kyoushi.simulation.transitions.Transition]
        originating from this state
        """
        return list(self._transitions.values())

    @property
    def transitions_map(self) -> Dict[str, Transition]:
        """
        List of all possible [`transitions`][cr_kyoushi.simulation.transitions.Transition]
        originating from this state
        """
        return self._transitions

    def __init__(
        self,
        name: str,
        transitions: List[Transition],
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The state name
            transitions: List of possible transitions
            name_prefix: A prefix for the state name

        Raises:
            ValueError: If there are transitions with duplicate names
        """
        self._name = name
        self._transitions = {t.name: t for t in transitions}
        self._name_prefix: Optional[str] = name_prefix

        if len(self._transitions) < len(transitions):
            raise ValueError("Transition names must be unique")

    @abstractmethod
    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        """Selects the next state transition.

           The selection logic depends on the state implementation. It might
           rely on the [`state machine context`][cr_kyoushi.simulation.sm.Statemachine.context]
           and/or execution environment information to select transitions based on complex conditions.

        Args:
            log: The bound logger initialized with transition specific information
            context (Context): State machine context which can be used for complex selection logic

        Returns:
            The selected [`Transition`][cr_kyoushi.simulation.transitions.Transition]
            or None if no transition is available.

        """
        ...

    def __str__(self):
        return f"name='{self.name}' transitions={self.transitions}"

    def __repr__(self):
        return f"{self.name}(name='{self.name}', transitions={self.transitions})"


class SequentialState(State):
    """Simple sequential state only having one possible transition"""

    def __init__(
        self,
        name: str,
        transition: Transition,
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The state name
            transition: The target transition
            name_prefix: A prefix for the state name

        Raises:
            ValueError: If transition is None
        """
        if transition is None:
            raise ValueError("Transition must not be None")

        super().__init__(name, [transition], name_prefix)
        self.__transition = transition

    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        return self.__transition


class ChoiceState(State):
    """Simple boolean choice state decides between two transitions"""

    @property
    def yes(self) -> Transition:
        """The transition that is returned when the decision function returns `True`"""
        return self.transitions[0]

    @property
    def no(self) -> Transition:
        """The transition that is returned when the decision function returns `False`"""
        return self.transitions[1]

    def __init__(
        self,
        name: str,
        decision_function: Callable[[BoundLogger, Context], bool],
        yes: Transition,
        no: Transition,
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The state name
            decision_function: Context function that decides a yes/no question.
            yes: The transition to return when the decision function returns `True`
            no: The transition to return when the decision function returns `False`
            name_prefix: A prefix for the state name
        """
        super().__init__(name, [yes, no], name_prefix)
        self.__decision_function: Callable[
            [BoundLogger, Context], bool
        ] = decision_function

    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        return self.yes if self.__decision_function(log, context) else self.no


class FinalState(State):
    """State with not further transitions which can be used as final state of a state machine"""

    def __init__(self, name: str, name_prefix: Optional[str] = None):
        """
        Args:
            name: The state name
            name_prefix: A prefix for the state name
        """
        super().__init__(name, [], name_prefix)

    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        return None


class RoundRobinState(State):
    def __init__(
        self,
        name: str,
        transitions: List[Transition],
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name (str): The state name
            transitions (List[Transition]): List of transitions to cycle through
            name_prefix: A prefix for the state name

        Raises:
            ValueError: If there are transitions with duplicate names
        """
        super().__init__(name, transitions, name_prefix)
        self.transition_cycle = cycle(transitions)

    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        try:
            return next(self.transition_cycle)
        except StopIteration:
            return None


class ProbabilisticState(State):
    """A state that uses a propability table to select its successor"""

    @property
    def weights(self) -> Sequence[float]:
        """The weight assigned to the transitions."""
        return self.__weights

    def __init__(
        self,
        name: str,
        transitions: List[Transition],
        weights: Sequence[float],
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The state name
            transitions: The list of transitions
            weights: The list of weights to assign to the transitions in probability notation.
            name_prefix: A prefix for the state name

        Raises:
            ValueError: If there are transitions with duplicate names
            ValueError: If the weights and transitions list lengths do not match
            ValueError: If the given weights do not sum up to 1
        """
        # initial base properties
        super().__init__(name, transitions, name_prefix)

        # convert weights to cumulative weights
        self.__weights = weights

        # verify that given weights and transitions are sound
        self.__verify_weights()

    def __verify_weights(self) -> None:
        # check that lengths match
        if len(self.transitions) != len(self.weights):
            raise ValueError(
                f"Size of transitions and weights do not match, \
                got transitions={len(self.transitions)} and weights={len(self.weights)}"
            )
        # if we were given an empty transition list
        # then there is nothing more to check
        if len(self.weights) > 0:
            # check that all probs are positive values
            if any(p < 0 for p in self.weights):
                raise ValueError("Probabilities cannot be negative!")
            # probabilities must always sum to 1
            if abs(1.0 - sum(self.weights)) > 1e-8:
                raise ValueError(
                    "Probabilities are uneven, sum of weights must be 1,"
                    f" but got {self.weights[-1]}!"
                )

    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        if len(self.transitions) > 0:
            return np.random.choice(a=np.array(self.transitions), p=self.weights)
        return None


class AdaptiveProbabilisticState(ProbabilisticState):
    """Special probabilistic state that allows modifaction of weights."""

    _modifiers: Dict[Transition, float]

    @property
    def modifiers(self) -> List[float]:
        """The weight modifiers assigned to the transitions."""
        return list(self._modifiers.values())

    @property
    def probabilities(self) -> List[float]:
        """The propabilities based on the weights and modifiers"""
        return calculate_propabilities(self.weights, self.modifiers)

    def __init__(
        self,
        name: str,
        transitions: List[Transition],
        weights: Sequence[float],
        modifiers: Optional[Sequence[float]] = None,
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The name of the state
            transitions: The list of transitions
            weights: The list of weights to assign to the transitions in propability notation
            modifiers: List of multiplicative modifiers for each weight.
                       Will default to all 1 if not set.
            name_prefix: A prefix for the state name
        """
        super().__init__(name, transitions, weights, name_prefix)
        if modifiers is None:
            modifiers = [1.0] * len(self.weights)

        self._modifiers = dict(zip(transitions, list(modifiers)))

        self.__modifiers_org: Tuple[float, ...] = tuple(self.modifiers)

    def adapt_before(self, log: BoundLogger, context: Context):
        """Hook to update the weight modifiers before the transition selection.

        Args:
            log: The logger for the sm context
            context: The state machine context
        """

    def adapt_after(
        self,
        log: BoundLogger,
        context: Context,
        selected: Optional[Transition],
    ):
        """Hook to update the weight modifiers after the transition selection.

        Args:
            log: The logger for the sm context
            context: The state machine context
            selected: The transition selected in this next call
        """

    def reset(self):
        """Resets the modifiers to their original state"""
        self._modifiers = dict(zip(self.transitions, self.__modifiers_org))

    def next(self, log: BoundLogger, context: Context) -> Optional[Transition]:
        if len(self.transitions) > 0:
            self.adapt_before(log, context)
            selected = np.random.choice(
                a=np.array(self.transitions), p=self.probabilities
            )
            self.adapt_after(log, context, selected)
            return selected
        return None


class EquallyRandomState(ProbabilisticState):
    """Special type of probabilistic state using an equal random distribution for all transitions"""

    def __init__(
        self,
        name: str,
        transitions: List[Transition],
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The state name
            transitions: The list of transitions
            name_prefix: A prefix for the state name

        Raises:
            ValueError: If there are transitions with duplicate names
        """
        # create even random distribution
        probability = 1.0 / len(transitions)
        weights = [probability for i in range(0, len(transitions))]
        # initialize using super
        super().__init__(name, transitions, weights, name_prefix)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cr_kyoushi', 'cr_kyoushi.simulation']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'colorama>=0.4.4,<0.5.0',
 'livereload>=2.6.3,<3.0.0',
 'numpy>=1.19.5,<2.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'ruamel.yaml>=0.16.12,<0.17.0',
 'structlog>=20.2.0,<21.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=3.1.0,<4.0.0',
                             'typing-extensions>=3.7.4,<4.0.0']}

entry_points = \
{'console_scripts': ['kyoushi-sim = kyoushi.simulation.cli:cli']}

setup_kwargs = {
    'name': 'kyoushi-simulation',
    'version': '0.3.7',
    'description': '',
    'long_description': '# AIT Cyber Range Kyoushi - Simulation\n\nThe Kyoushi Simulation package provides a development API and cli utilities for creating\nand running state machines. Developed state machines can be executed to automate the simulation\nof both attacker and normal user behavior in a Cyber Range. Thus facilitating Cyber Range\nexercises, IDS data set generation and other Cyber Range related tasks.\n',
    'author': 'Maximilian Frank',
    'author_email': 'maximilian.frank@ait.ac.at',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git-service.ait.ac.at/sct-cyberrange/tools/kyoushi/simulation',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Dict,
    Optional,
)

import click

from . import plugins
from .config import (
    LogLevel,
    Settings,
    configure_seed,
    load_settings,
    load_sm_config,
)
from .logging import (
    configure_logging,
    get_logger,
)


try:
    from importlib.metadata import EntryPoint  # type: ignore
except ImportError:
    # need to use backport for python < 3.8
    from importlib_metadata import EntryPoint

__all__ = ["CliPath", "Info", "cli", "version", "sm_list", "sm_run"]

logger = get_logger()


class EnumChoice(click.Choice):
    """Click choice from an enum

    from https://github.com/pallets/click/issues/605#issuecomment-582574555
    """

    case_sensitive: bool
    use_value: bool

    def __init__(self, enum, case_sensitive=False, use_value=False):
        self.enum = enum
        self.use_value = use_value
        choices = [
            str(e.value) if use_value else e.name
            for e in self.enum.__members__.values()
        ]
        super().__init__(choices, case_sensitive)

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ):
        if isinstance(value, Enum) and value in self.enum:
            return value
        result = super().convert(value, param, ctx)

        # if we got here result must be a value or enum key
        # since super will convert one of the choices

        if self.use_value:
            # using list instead of generator for test coverage
            # also we don't care about efficiency to much here
            return [e for e in self.enum if str(e.value) == result][0]

        return self.enum[result]


class CliPath(click.Path):
    """A Click path argument that returns a pathlib Path, not a string"""

    def convert(self, value, param, ctx):
        return Path(super().convert(value, param, ctx))


class Info:
    """An information object to pass data between CLI functions."""

    def __init__(self):  # Note: This object must have an empty constructor.
        """Create a new instance."""
        self.settings_path: Optional[Path] = None
        self.settings: Optional[Settings] = None
        self.available_factories: Optional[Dict[str, EntryPoint]] = None


# pass_info is a decorator for functions that pass 'Info' objects.
#: pylint: disable=invalid-name
pass_info = click.make_pass_decorator(Info, ensure=True)


# Change the options to below to suit the actual options for your task (or
# tasks).
@click.group()
@click.option("--log-level", type=EnumChoice(LogLevel), help="The log level")
@click.option(
    "--seed",
    default=None,
    type=click.INT,
    help="Global seeds for PRNGs used during simulation",
)
@click.option(
    "--config",
    "-c",
    type=CliPath(dir_okay=False, readable=True),
    default="config.yml",
    show_default=True,
    help="The Cyber Range Kyoushi Simulation settings file",
)
@pass_info
def cli(info: Info, log_level: LogLevel, seed: Optional[int], config: Path):
    """Run Cyber Range Kyoushi Simulation."""
    info.settings_path = config
    info.settings = load_settings(info.settings_path, log_level=log_level, seed=seed)
    info.available_factories = plugins.get_factories(info.settings.plugin)

    # setup logging
    configure_logging(info.settings.log)

    # setup prngs
    configure_seed(info.settings.seed)


@cli.command()
@pass_info
def version(info: Info):
    """Get the library version."""
    from .util import version_info

    click.echo(version_info(cli_info=info))


@cli.command(name="list")
@pass_info
def sm_list(info: Info):
    """List available state machine factories."""
    assert info.available_factories is not None
    click.echo(click.style("Available state machine factories:", bold=True))
    for factory_plugin in info.available_factories.values():
        click.echo(f"\t{factory_plugin.name} - {factory_plugin.value}")


@cli.command(name="run")
@pass_info
@click.option(
    "--sm-config",
    "-s",
    type=CliPath(dir_okay=False, readable=True),
    default="sm.yml",
    show_default=True,
    help="The state machine configuration file",
)
@click.option(
    "--factory",
    "-f",
    type=str,
    required=True,
    help="""
        The state machine factory to use.
        This can be either the name of a statemachine factory entrypoint plugin or
        the path to a python file containing a statemachine factory.
        """,
)
def sm_run(info: Info, sm_config: Path, factory: str):
    """Execute a state machine."""
    assert info.available_factories is not None

    # get factory
    factory_obj = plugins.get_factory(info.available_factories, factory)
    StatemachineConfig = factory_obj.config_class

    # load state machine config and build machine
    config = load_sm_config(sm_config, StatemachineConfig)
    logger.debug("Loaded config", config=config)
    machine = factory_obj.build(config)

    # execute machine
    machine.run()



# -*- coding: utf-8 -*-
"""Statemachine util module

This module contains some utility functions which can used for statemachines, state
and state transitions.
"""
from __future__ import annotations

import signal
import time

from contextlib import contextmanager
from datetime import (
    datetime,
    tzinfo,
)
from types import FrameType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    Union,
)

from .errors import SkipSectionError
from .logging import get_logger
from .model import ApproximateFloat


if TYPE_CHECKING:
    from .cli import Info


__all__ = ["version_info", "elements_unique", "skip_on_interrupt", "sleep"]

log = get_logger()


def version_info(cli_info: Info) -> str:
    """Returns formatted version information about the `cr_kyoushi.simulation package`.

    Adapted from
    [Pydantic version.py](https://github.com/samuelcolvin/pydantic/blob/master/pydantic/version.py)
    """
    import platform
    import sys

    from pathlib import Path

    from . import __version__

    info = {
        "cr_kyoushi.simulation version": __version__,
        "config path": cli_info.settings_path.absolute()
        if cli_info.settings_path is not None
        else None,
        "install path": Path(__file__).resolve().parent,
        "python version": sys.version,
        "platform": platform.platform(),
    }
    return "\n".join(
        "{:>30} {}".format(k + ":", str(v).replace("\n", " ")) for k, v in info.items()
    )


def elements_unique(to_check: List[Any]) -> bool:
    """Utility function to check if elements in a list are uniq.

    Args:
        to_check: The list to check

    Returns:
        `True` if the list contains no duplicates `False` otherwise.
    """
    seen = set()
    return not any(i in seen or seen.add(i) for i in to_check)  # type: ignore


def normalize_propabilities(
    propabilities: Sequence[float], check_positive: bool = True
) -> List[float]:
    """Normalizes a propability distribution to sum up to 1.

    Has a floating point error tolerance of up to 1e-8 (same as numpy.choice).
    If the resulting difference is greater than 1e-8 it is added to the first
    non zero propability.

    Args:
        propabilities: The distribution to normalize
        check_positive: Switch to disable the positive number test.
                        This makes the function slightly faster

    Raises:
        ValueError: If the distribution sums to 0 or
                    If one of the propability values is negative

    Returns:
        List[float]: [description]
    """
    # only check when requested
    if check_positive and any(p < 0 for p in propabilities):
        raise ValueError("Propabilities must be positive numbers")

    total = sum(propabilities)
    if total > 0:
        multiplier = 1.0 / total
        propabilities = [p * multiplier for p in propabilities]
        diff = 1.0 - sum(propabilities)
        # when rounding errors become to extrem
        # then we fix them by adding the diff to 1 a propability
        if abs(diff) > 1e-8:
            # first non 0 propability index will be increased
            adjust_index = next((i for i, x in enumerate(propabilities) if x != 0.0))
            propabilities[adjust_index] += diff
        return propabilities

    raise ValueError("Resulting propabilities sum to 0")


def calculate_propabilities(
    weights: Sequence[float],
    modifiers: Sequence[float],
) -> List[float]:
    """Calculates the propability distribution from a list of weights and modifiers.

    Args:
        weights: The base weights
        modifiers: The weight modifiers

    Raises:
        ValueError: If the len of weights and modifiers do not match
                    If the resulting propabilities do not sum to 1

    Returns:
        The modified weights as propabilities
    """
    if len(weights) != len(modifiers):
        raise ValueError("The len of weights and modifieres do not match!")

    if abs(1.0 - sum(weights)) > 1e-8:
        raise ValueError("The weights must sum up to 1")

    if any(w < 0 for w in weights) or any(m < 0 for m in modifiers):
        raise ValueError("Weights and modifiers must be positive values")

    return normalize_propabilities(
        [w * m for w, m in zip(weights, modifiers)],
        # weights[*] and modifiers[*] > 0 already guarantees this
        check_positive=False,
    )


def skip_on_interrupt_sig_handler(signum: signal.Signals, frame: FrameType):
    """Simple signal handler that just raises a
    [`SkipSectionError`][cr_kyoushi.simulation.errors.SkipSectionError]
    when it receives a interrupt signal.

    Raises:
        SkipSectionError: To indicate the current section should be skipped
    """
    log.debug("Received interrupt raising skip section error")
    raise SkipSectionError()


@contextmanager
def skip_on_interrupt(
    sig: signal.Signals = signal.SIGINT,
    sig_handler: Callable[
        [signal.Signals, FrameType], None
    ] = skip_on_interrupt_sig_handler,
):
    """Context manager for creating skipable code sections.

    Args:
        sig: The signal that should indicate a skip request.
        sig_handler: The skip on signal handler to use.

    Example:
        ```python
            # start of the skipable section
            with skip_on_interrupt():
                ...
                long_running_task()
                ...
            # after a skip we resume here
            ...
        ```
    """
    try:
        original_handler = signal.getsignal(sig)
        signal.signal(sig, sig_handler)
        yield
    except SkipSectionError:
        log.debug("Skipped section")
        signal.signal(sig, original_handler)
        # clear original signal handler so we know not to
        # reset to it twice
        original_handler = None
        print("Press CTRL+C again to stop the program")
        time.sleep(0.5)
    finally:
        if original_handler is not None:
            signal.signal(sig, original_handler)


def now(tz: Optional[tzinfo] = None) -> datetime:
    """Utility function for getting the current datetime.

    For statemachine features that require the current time it is preferred
    to use this function instead of `datetime.now()` directly.
    This suggested because `datetime` is a builtin and a such cannot
    easily be mocked during tests.

    Args:
        tz: Optionally the timezone to use
    """
    return datetime.now(tz)


def sleep(sleep_time: Union[ApproximateFloat, float]) -> None:
    """Skipable sleep function

    This function utilizes the
    [`skip_on_interrupt`][cr_kyoushi.simulation.util.skip_on_interrupt]
    context manager to implement a skipable sleep.

    `SIGINT` is used used as skip signal. For CLI applications
    simply press ++ctrl+c++ to skip to interrupt the sleep.

    !!! Note
        If you wish to send the `SIGINT` signal to the main process press
         ++ctrl+c++ twice.
    """
    if isinstance(sleep_time, ApproximateFloat):
        sleep_time = sleep_time.value

    with skip_on_interrupt():
        log.debug("Going to sleep for %f", sleep_time)
        time.sleep(sleep_time)
        log.debug("Resuming execution after sleeping for %f", sleep_time)


def sleep_until(
    end_datetime: datetime,
    min_sleep_amount: float = 0.1,
    sleep_amount: Optional[Union[float, ApproximateFloat]] = None,
):
    """Sleep until specified datetime

    The default behavior is to basically binary search towards the target datetime.
    i.e., the sleep duration is always `time left/2` until min sleep amount is larger
    than the division result. Alternatively fixed sleep steps can be configured.

    !!! Hint
        You can interupt the current sleep and check if the desired datetime is already
        reached by pressing ++ctrl+c++ and sending a `SIGINT`.


    Args:
        end_datetime: The datetime to wait until
        min_sleep_amount: The minimum amount of time sleep in between checks.
        sleep_amount: Optionally use fixed amount of time sleep steps
    """

    while True:
        diff = (end_datetime - now()).total_seconds()

        # stop waiting once its the end time or later
        if diff <= 0:
            return

        # if we do not have a sleep interval sleep relative to
        # the time between now and the end time
        if sleep_amount is None:
            #  to avoid many extremely short sleeps
            #  fallback to the minimum sleep
            sleep(max(min_sleep_amount, diff / 2))
        else:
            sleep(sleep_amount)



import sys

from types import FunctionType
from typing import (
    Callable,
    Optional,
    Union,
)

from structlog.stdlib import BoundLogger


if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    from typing_extensions import Protocol

from .model import (
    ApproximateFloat,
    Context,
)
from .util import sleep


__all__ = [
    "Transition",
    "DelayedTransition",
    "transition",
    "delayed_transition",
    "NoopTransition",
    "noop",
]


class TransitionFunction(Protocol):
    def __call__(
        self,
        log: BoundLogger,
        current_state: str,
        context: Context,
        target: Optional[str],
    ):
        """

        Args:
            log: The bound logger initialized with transition specific information
            current_state: The calling states name
            context: The state machine context
            target: The target state

        Raises:
            TransitionExecutionError: If a transition error occurs for which we can fallback into a valid state
        """


class Transition:
    """Abstract base Transition class describes a transition from one state into another"""

    @property
    def name(self) -> str:
        """The name of the transition (including the prefix)"""
        if self._name_prefix is not None:
            return f"{self._name_prefix}_{self._name}"
        return self._name

    @property
    def name_only(self) -> str:
        """The name of the transition"""
        return self._name

    @property
    def name_prefix(self) -> Optional[str]:
        """The name prefix of the transition instance."""
        return self._name_prefix

    @property
    def target(self) -> Optional[str]:
        """The target state of the transition"""
        return self._target

    @property
    def transition_function(self) -> TransitionFunction:
        """The transition function"""
        return self._transition_function

    def __init__(
        self,
        transition_function: TransitionFunction,
        name: Optional[str] = None,
        target: Optional[str] = None,
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            transition_function: The transition function to call upon execution
            name: The transition name
            target: The target state
            name_prefix: A prefix for the transition name
        """
        if name is None:
            if isinstance(transition_function, FunctionType):
                name = transition_function.__name__.lower()
            else:
                name = transition_function.__class__.__name__.lower()

        self._transition_function: TransitionFunction = transition_function
        self._name: str = name
        self._name_prefix: Optional[str] = name_prefix
        self._target: Optional[str] = target

    def execute(
        self,
        log: BoundLogger,
        current_state: str,
        context: Context,
    ) -> Optional[str]:
        """Transition execution function called by the state machine.

        The default behavior is to directly call the configured
        [`transition_function`][cr_kyoushi.simulation.transitions.Transition.transition_function]


        !!! Info
            This function can be overridden or extended to change **how**
            transitions are executed. Actual transition implementations
            should be done via the supplied
            [`transition_function`][cr_kyoushi.simulation.transitions.Transition.transition_function].

        !!! Warning
            Don't forget to call `#!python super().execute(current_state, context)`
            or `#!python self.transition_function(current_state, context)`.

        Args:
            log: The bound logger initialized with transition specific information
            current_state: The calling states name
            context: The state machine context

        Returns:
            The state the machine has moved to

        Raises:
            TransitionExecutionError: If a transition error occurs for which we can fallback into a valid state
        """
        self.transition_function(log, current_state, context, self.target)
        return self.target

    def __str__(self):
        return f"name='{self.name}' -> target={self.target}"

    def __repr__(self):
        return f"{self.name}(name='{self.name}', target={self.target})"


def transition(
    name: Optional[str] = None,
    target: Optional[str] = None,
    name_prefix: Optional[str] = None,
) -> Callable[[TransitionFunction], Transition]:
    """Transition decorator that can be used to turn a
    [`TransitionFunction`][cr_kyoushi.simulation.transitions.TransitionFunction] into a
    [`Transition`][cr_kyoushi.simulation.transitions.Transition].

    Args:
        name: The name of the transition
        target: The target states name
        name_prefix: A prefix for the transition name


    Example:
        ```python
            @transition(name="example", target="next")
            def example(current_state, context, target):
                ...

            # the above is equivalent to
            def example_func(current_state, context, target):
                ...

            example = Transition(example_func, name="example", target="next")
        ```

    Returns:
        A decorator that turns a [`TransitionFunction`][cr_kyoushi.simulation.transitions.TransitionFunction]
        into a Transition initialized with the given args.
    """

    def decorator(func: TransitionFunction) -> Transition:
        return Transition(
            transition_function=func, target=target, name=name, name_prefix=name_prefix
        )

    return decorator


class DelayedTransition(Transition):
    """
    Abstract DelayedTransition allows configuring skipable
    pre and post transition execution delays.
    """

    @property
    def delay_before(self) -> ApproximateFloat:
        """The amount of time in seconds to delay before execution"""
        return self._delay_before

    @property
    def delay_after(self) -> ApproximateFloat:
        """The amount of time in seconds to delay after execution"""
        return self._delay_after

    def __init__(
        self,
        transition_function: TransitionFunction,
        name: Optional[str] = None,
        target: Optional[str] = None,
        delay_before: Union[ApproximateFloat, float] = 0.0,
        delay_after: Union[ApproximateFloat, float] = 0.0,
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The transition name
            transition_function: The transition function to call upon execution
            target: The target state
            delay_before: The pre execution delay to configure
            delay_after: The post execution delay to configure
            name_prefix: A prefix for the transition name
        """
        super().__init__(transition_function, name, target, name_prefix)

        if isinstance(delay_before, float):
            delay_before = ApproximateFloat.convert(delay_before)

        if isinstance(delay_after, float):
            delay_after = ApproximateFloat.convert(delay_after)

        self._delay_before = delay_before
        self._delay_after = delay_after

    def execute(
        self,
        log: BoundLogger,
        current_state: str,
        context: Context,
    ) -> Optional[str]:
        """
        Delayed transition execution sleeps before and after executing the transition.

        Both delays can be configured during initialization.
        The delays use a special [`sleep`][cr_kyoushi.simulation.util.sleep] function that registers a `SIGINT`
        signal handler to make it possible to interrupt and skip the sleep
        phase.

        !!! Note
            When running in a terminal press ++ctrl+c++ to
            *fast forward*.

        Args:
            log: The bound logger initialized with transition specific information
            current_state: The calling states name
            context: The state machine context

        Returns:
            The state the machine has moved to

        Raises:
            TransitionExecutionError: If a transition error occurs for which we can fallback into a valid state
        """
        sleep(self.delay_before)
        next_state = super().execute(log, current_state, context)
        sleep(self.delay_after)
        return next_state


def delayed_transition(
    name: Optional[str] = None,
    target: Optional[str] = None,
    delay_before: Union[ApproximateFloat, float] = 0.0,
    delay_after: Union[ApproximateFloat, float] = 0.0,
    name_prefix: Optional[str] = None,
) -> Callable[[TransitionFunction], DelayedTransition]:
    """Transition decorator that can be used to turn a
    [`TransitionFunction`][cr_kyoushi.simulation.transitions.TransitionFunction] into a
    [`DelayedTransition`][cr_kyoushi.simulation.transitions.DelayedTransition].

    Args:
        name: The name of the transition
        target: The target states name
        delay_before: The pre execution delay to configure
        delay_after: The post execution delay to configure
        name_prefix: A prefix for the transition name

    Example:
        ```python
            @delayed_transition(name="example", target="next", delay_before=1.5)
            def example(current_state, context, target):
                ...

            # the above is equivalent to
            def example_func(current_state, context, target):
                ...

            example = DelayedTransition(example_func, name="example", target="next", delay_before=1.5)
        ```

    Returns:
        A decorator that turns a [`TransitionFunction`][cr_kyoushi.simulation.transitions.TransitionFunction]
        into a DelayedTransition initialized with the given args.
    """

    def decorator(func: TransitionFunction) -> DelayedTransition:
        return DelayedTransition(
            transition_function=func,
            target=target,
            name=name,
            delay_before=delay_before,
            delay_after=delay_after,
            name_prefix=name_prefix,
        )

    return decorator


def noop(log: BoundLogger, current_state: str, context: Context, target: Optional[str]):
    """No operation transition function"""


class NoopTransition(Transition):
    """No noperation transition that only changes the current state."""

    def __init__(
        self,
        name: str = "noop",
        target: Optional[str] = None,
        name_prefix: Optional[str] = None,
    ):
        """
        Args:
            name: The name of the transition
            target: The name of the target state
            name_prefix: A prefix for the transition name
        """
        super().__init__(noop, name=name, target=target, name_prefix=name_prefix)



# -*- coding: utf-8 -*-
"""State machine module

This module contains all class and function defintions for creating and defining
Cyber Range Kyoushi simulation machines.
"""
from abc import (
    ABC,
    abstractmethod,
)
from datetime import datetime
from typing import (
    Dict,
    Generic,
    List,
    Optional,
    Type,
)
from uuid import (
    UUID,
    uuid4,
)

from structlog.stdlib import BoundLogger

from . import errors
from .config import get_seed
from .logging import get_logger
from .model import (
    Context,
    StatemachineConfig,
    WorkSchedule,
)
from .states import State
from .transitions import Transition
from .util import (
    now,
    sleep_until,
)


__all__ = ["Statemachine", "StatemachineFactory"]


class Statemachine:
    """
    Implements state control and transition logic

    This class implements basic state machine execution, i.e., execution
    starts at the configured initial state and continues until a end state
    (i.e., a state without outgoing transitions) is reached. A state machine
    can be started by calling the [`run()`][cr_kyoushi.simulation.sm.Statemachine.run] function.

    !!! Note
        You can also execute a state machine step wise using `execute_step()`
        function manually. If you choose to do so be care full to remember to call
        `setup_context()` before you start execution and `destroy_context()`
        after the state machine is finished.

    Default state machine behavior can be extended or modified by creating
    a sub class and overriding the state machine functions
    [`execute_machine()`][cr_kyoushi.simulation.sm.Statemachine.execute_machine],
    [`execute_step()`][cr_kyoushi.simulation.sm.Statemachine.execute_step],
    [`execute_transition()`][cr_kyoushi.simulation.sm.Statemachine.execute_transition], etc.
    """

    @property
    def context(self) -> Context:
        """The state machine execution context object."""
        return self._context

    @context.setter
    def context(self, new_value: Context):
        self._context = new_value

    @property
    def log(self) -> BoundLogger:
        """Bound logger initialize with context information for this state machine."""
        return self.__log

    @property
    def uuid(self) -> UUID:
        """UUID for this state machine execution.

        For example used as unique run ID during when logging.
        """
        return self.__uuid

    def __init__(
        self,
        initial_state: str,
        states: List[State],
        max_errors: int = 0,
    ):
        """
        Args:
            initial_state: The name of the initial state
            states: List of all states the state machine can enter
            max_errors: Maximum amount of errors the state machine is allowed to encounter
                        before it stops trying to recover by reseting to the initial state.

        """
        self.initial_state = initial_state
        self.current_state: Optional[str] = initial_state
        self.current_transition: Optional[Transition] = None
        self.states: Dict[str, State] = {state.name: state for state in states}
        self.context: Context = {}
        self.max_errors = max_errors
        self.errors = 0
        self.__uuid = uuid4()
        self.__log: BoundLogger = get_logger().bind(run=self.uuid)
        self.log.info("Created state machine", seed=get_seed())

    def setup_context(self) -> None:
        """Initialize and setup the state machine execution context

        ??? Note
            Override this function if your state machine needs run some
            logic or set `Context` information before it can be executed.
        """

    def destroy_context(self) -> None:
        """Destroy and clean up the state machine execution context

        ??? Note
            Override or extend this function if your state machine needs run some
            logic or free `Context` information after it has finished executing.
        """

    def execute_transition(self, log: BoundLogger):
        """Execute the current transition.

        The current transition is executed and the current state is
        updated on successful executions. This function also handles
        [`TransitionExecutionErrors`][cr_kyoushi.simulation.errors.TransitionExecutionError]
        and sets the state machines current state to the errors fallback state.

        Args:
            log: The bound logger initialized with transition specific information

        ??? Note
            Override or extend this if you want to change how transitions are executed
            or how [`TransitionExecutionErrors`][cr_kyoushi.simulation.errors.TransitionExecutionError]
            are handled.
        """
        assert self.current_state is not None
        assert self.current_transition is not None
        try:
            log.info(
                "Executing transition %s -> %s",
                self.current_state,
                self.current_transition,
            )
            self.current_state = self.current_transition.execute(
                log, self.current_state, self.context
            )
            log.info("Moved to new state", new_state=self.current_state)
        except errors.TransitionExecutionError as transition_error:
            log.warning("Encountered a transition error: %s", transition_error)
            if transition_error.fallback_state:
                log.warning(
                    "Recovering to fallback state",
                    fallback=transition_error.fallback_state,
                )

            self.current_state = transition_error.fallback_state

    def execute_step(self):
        """Execute a single state machine step.

        This function delegates transition execution to `execute_transition()`.
        All pre and post execution tasks such as retrieving the transition
        from the current state before the transition and handeling unexpected
        errors encountered during transition execution.

        ??? Note
            Override or extend this function if you whish to change pre-, post-execution
            and handling of all unexpected errors.
        """
        assert self.current_state is not None

        # bind upcoming transition context to logger
        log: BoundLogger = self.log.bind(
            current_state=self.current_state,
            transition=None,
            transition_id=uuid4(),
            target=None,
        )

        try:
            self.current_transition = self.states[self.current_state].next(
                log, self.context
            )

            if self.current_transition:
                # bind selected transition and target to logger
                log = log.bind(
                    transition=self.current_transition.name,
                    target=self.current_transition.target,
                )

                # execute transition
                self.execute_transition(log)
            else:
                log.info("Empty transition received state machine will end")
                self.current_state = None
        except Exception:
            log.exception("State machine execution failure")

            # try to recover from error by restarting state machine
            self.errors += 1
            if self.max_errors > self.errors:
                self.destroy_context()
                log.warning("Trying to recover statemachine from exception ")
                self.setup_context()
                self.current_state = self.initial_state
            else:
                self.current_state = None

    def execute_machine(self):
        """State machine main execution loop.

        This function executes state machine steps in a loop
        until a end state is reached (i.e., current state is `None`).

        ??? Note
            Override or extends this if you whish to change how your
            state machine does continues execution.
        """
        # state machine run main loop
        while self.current_state:
            self.execute_step()

    def run(self) -> None:
        """Starts the state machine execution.

        The state machine execution context is setup before
        executing the state machine main loop and destroyed again
        after the main loop ends.
        """
        # prepare state machine before start
        self.log.info("Starting state machine")
        self.setup_context()

        # execute the state machine
        self.log.info("Entering state machine execution")
        self.execute_machine()

        # clean up state machine
        self.destroy_context()
        self.log.info("State machine finished")


class StartEndTimeStatemachine(Statemachine):
    """Special type of state machine that allows configuring a start and end time.

    The main stat machine execution loop will only start execution once it is
    the configured start time (starts immediately if none is configured). Similarly
    the state machine will stop once the current time is the end time even if there
    are still transitions left to execute. This is for example useful when you wish to
    configure a state machine to only run for the duration of an experiment.

    !!! Note
        A state machine might end before its end time if it enters a final state.
    """

    @property
    def start_time(self) -> Optional[datetime]:
        """The `datetime` this state machine will start execution"""
        return self.__start_time

    @property
    def end_time(self) -> Optional[datetime]:
        """The `datetime` this state machine will end"""
        return self.__end_time

    def __init__(
        self,
        initial_state: str,
        states: List[State],
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        max_errors: int = 0,
    ):
        """
        Args:
            initial_state: The name of the initial state
            states: List of all states the state machine can enter
            start_time: The `datetime` this state machine should start execution
            end_time: The `datetime` this state machine should end execution
            max_errors: Maximum amount of errors the state machine is allowed to encounter
                        before it stops trying to recover by reseting to the initial state.

        !!! Note
            If `state_time` or `end_time` are `None` then the state machine will
            start or end execution normally.
        """
        super().__init__(initial_state, states, max_errors=max_errors)
        self.__start_time = start_time
        self.__end_time = end_time

    def _is_end_time(self) -> bool:
        # if no end time was set then this is always false
        if self.end_time is None:
            return False
        return self.end_time <= now()

    def execute_machine(self):
        """State machine main execution loop.

        This function executes state machine steps in a loop until either
         - a end state is reached (i.e., current state is `None`)
         - or the current time is >= `end_time`

        """
        # state machine run main loop
        while self.current_state and not self._is_end_time():
            self.execute_step()

    def run(self):
        """Starts the state machine execution.

        This will only start the state machine execution
        once the given start time is reached.
        """
        # wait for start time before actually starting the machine
        if self.start_time is not None:
            sleep_until(self.start_time)
        return super().run()


class WorkHoursStatemachine(StartEndTimeStatemachine):
    """State machine optionally allows the configuration of work hours.

    !!! Note
        This state machine extends
        [`StartEndTimeStatemachine`][cr_kyoushi.simulation.sm.StartEndTimeStatemachine]
        and as such has all its features.

    Work hours are defined through the configuration of a [`ActivePeriod`][cr_kyoushi.simulation.model.ActivePeriod].
    Outside of its work hours this state machine will simply idle and do nothing. You can also configure
    """

    @property
    def work_schedule(self) -> Optional[WorkSchedule]:
        """The work schedules for this state machine"""
        return self.__work_schedule

    def __init__(
        self,
        initial_state: str,
        states: List[State],
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        work_schedule: Optional[WorkSchedule] = None,
        max_errors: int = 0,
    ):
        """
        Args:
            initial_state: The name of the initial state
            states: List of all states the state machine can enter
            start_time: The `datetime` this state machine should start execution
            end_time: The `datetime` this state machine should end execution
            work_schedule: The state machines work schedule (days and times to work)
            max_errors: Maximum amount of errors the state machine is allowed to encounter
                        before it stops trying to recover by reseting to the initial state.

        !!! Note
            If `state_time` or `end_time` are `None` then the state machine will
            start or end execution normally.
        """
        super().__init__(
            initial_state,
            states,
            start_time=start_time,
            end_time=end_time,
            max_errors=max_errors,
        )
        self.__work_schedule = work_schedule

    def _in_work_hours(self) -> bool:
        # if we do not have work hours we work 24/7
        if self.work_schedule is None:
            return True

        # if we have work hours we have to check them
        return self.work_schedule.is_work_time(now())

    def _pause_work(self):
        """The pause work metho will be called before pausing the SM until the next work time.

        Use this method to destroy any resources that should not be active outside
        the work hours (e.g., the selenium browser). Also see `_resume_work`

        !!! Hint
            You could simply destroy your context and set the SM to the initial state:
            ```python
            self.current_state = self.initial_state
            # reset context
            self.destroy_context()
            ```
        """

    def _resume_work(self):
        """The resume work method will be called before resuming work after sleeping.

        Use this method to prepare the state machine to resume after a potentially long
        pause. By default this method does nothing. Also see `_pause_work`.

        !!! Hint
            You could for example configure your state machine to recreate from the initial
            state before resuming work:

            ```python
            self.setup_context()
            ```
        """

    def _wait_for_work(self):
        """Idle until it is time to work again.

        Before returning to the normal state machine flow this
        will also call [`_resume_work`][cr_kyoushi.simulation.sm.WorkHoursStatemachine._resume_work].

        !!! Note
            If the next potential work time is after the machines end time
            it will not sleep, but instead set the current state to `None`
            and let the state machine flow end execution.
        """
        # immediately return if there is no work schedule
        if self.work_schedule is None:
            return

        next_work = self.work_schedule.next_work_start(now())

        # if there is no next work time or the machine will end
        # before the next work we stop immediately
        if next_work is None or (
            self.end_time is not None and self.end_time <= next_work
        ):
            self.current_state = None
        else:
            self.log.info("Pausing state machine")
            self._pause_work()
            self.log.info("Paused state machine")
            # wait til we have work again
            sleep_until(next_work)
            # and then pre-pare to resume work
            self.log.info("Resuming state machine")
            self._resume_work()
            self.log.info("Resumed state machine")

    def execute_step(self):
        """Execute a single state machine step.

        This will only execute a step if the current time is within
        our work schedule. Outside the work time the state machine will
        [wait until work][cr_kyoushi.simulation.sm.WorkHoursStatemachine._wait_for_work] begins again.
        """
        # when we are in work hours business as usual
        if self._in_work_hours():
            super().execute_step()
        # outside of work hours we idle
        else:
            self._wait_for_work()


class StatemachineFactory(ABC, Generic[StatemachineConfig]):
    """Abstract class definition for factories generating state machines

    State machine factories are used by the CLI system to load dynamically load
    state machines from [entrypoints](https://packaging.python.org/specifications/entry-points/)
    or python files.

    A state machine factory must have a **name** and a **config class**.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the state machine factory."""

    @property
    @abstractmethod
    def config_class(self) -> Type[StatemachineConfig]:
        """The config class to use for the state machine.

        You can use the configuration class to define all required
        and optional configuration options for your state machine.
        e.g., to make it possible for users of your state machine to configure
        the probabilities for [`ProbabilisticStates`][cr_kyoushi.simulation.states.ProbabilisticState].

        The CLI system dynamically loads and validates configuration for your
        state machine based on the config class.
        [Pydantic](https://pydantic-docs.helpmanual.io/) is used for this so it is
        recommended to define your config class must be a pydantic model or any other field type
        pydantic can handle.
        """

    @abstractmethod
    def build(self, config: StatemachineConfig) -> Statemachine:
        """Builds the state machine instance.

        The build function must also create and initialize all states and transitions
        required by the state machine. This is called by the CLI system to create
        the state machine before executing it.

        Args:
            config: Configuration for your state machine.

        Returns:
            Statemachine: Statemachine instances created based on the given configuration.
        """
